#############################################################
#                                                           #
# Blender Domical Vault Script                              #
#                                                           #
# (C) December 2003 Neil McAllister                         #
#                                                           #
#############################################################
# History                                                   #
# V: 0.0.1 - 27-12-02 - First Working Version               #
#############################################################

import Blender
from Blender import *
from Blender.Draw import *
from math import *

BL_VERSION = Blender.Get('version')
if (BL_VERSION<=223):
	import Blender210

pointy=Create(0.5)
length=Create(4.0)
breadth=Create(6.0)
  
#############################################################
# MAIN                                                      #
#############################################################
def ConstructVault():
	global pointy, length, breadth
	me=NMesh.GetRaw()

	i=0.0
	j=0.0

	n=32.0
	l=length.val
	b=breadth.val
	p=pointy.val
	d=sqrt(l*l+b*b)
	alpha=acos(1-1/(2*p))
	theta=alpha/n
	rl=l*p*2
	rb=b*p*2
	rd=d*p*2

	for i in range(0,n+1):
		for j in range(0,n+1):
			x=(cos(alpha)-cos(theta*i))*rb
			y=(cos(alpha)-cos(theta*j))*rl
			if i<j:
				z=sin(theta*i)*(rd-rl)+sin(theta*j)*rl
			else:
				z=sin(theta*j)*(rd-rb)+sin(theta*i)*rb
			v=NMesh.Vert(x,y,z)
			me.verts.append(v)

	n0=len(range(0,n+1))

	for i in range(0,n):  
		for j in range(0,n):  
			if i==j:
				f=NMesh.Face()  
				f.v.append(me.verts[i*n0+j])  
				f.v.append(me.verts[i*n0+j+1])
				f.v.append(me.verts[(i+1)*n0+j+1])  
				me.faces.append(f)

				f=NMesh.Face()  
				f.v.append(me.verts[i*n0+j])  
				f.v.append(me.verts[(i+1)*n0+j+1])  
				f.v.append(me.verts[(i+1)*n0+j]) 
				me.faces.append(f) 

			else:
				f=NMesh.Face()  
				f.v.append(me.verts[i*n0+j])  
				f.v.append(me.verts[i*n0+j+1])  
				f.v.append(me.verts[(i+1)*n0+j+1])  
				f.v.append(me.verts[(i+1)*n0+j]) 
				me.faces.append(f) 

	NMesh.PutRaw(me,"plane",1)  
	Blender.Redraw()

#############################################################
# Graphics                                                  #
#############################################################
def Warn():
	BGL.glRasterPos2d(115, 23)
	Blender.Window.Redraw(Blender.Window.Const.TEXT)
	
def draw():
	global pointy, length, breadth
	BGL.glClearColor(0.5, 0.7, 0.5, 1)
	BGL.glColor3f(1,1,1)
	BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
	BGL.glColor3f(1, 1, 0)
	BGL.glRasterPos2d(8, 153)
	Text("Blender Domical Vault Script")
	BGL.glRasterPos2d(8, 133)
	Text("(C) Dec. 2003 Neil McAllister")
  
	length=Slider("Length ",2,10, 100, 200, 20, length.val, 0, 100)
	breadth=Slider("Breadth ",2,10, 75, 200, 20, breadth.val, 0, 100)
	pointy=Slider("Pointiness ",2,10, 50, 200, 20, pointy.val, 0.5, 4)
	Button("Exit", 1, 100, 10, 40, 20)

   
def event(evt, val): 
	if (evt== QKEY and not val):
		Exit()

def bevent(evt):
	if   (evt== 1):
		Exit()
	elif (evt== 2):
		ConstructVault()
		Draw()
	elif (evt== 3):
		ConstructVault()
		Redraw()

ConstructVault()
Register(draw, event, bevent)
